@@ -103,10 +103,10 @@ PipelineInfo = TypedDict('PipelineInfo', {
     "company": str,
     "high_priority": bool,
     "ins": List[str],
-    "is_parallel": bool,
     "name": str,
     "nodes": List[NodeInfo],
     "outs": List[Tuple[str, str]],
+    "queue_mng": Optional[str],
     "settings": Dict[str, Any],
     "state": str,
 })
@@ -137,7 +137,7 @@ PipelineDef = TypedDict('PipelineDef', {
     "nodes": List[NodeDef],
     "state": str,
     "high_priority": bool,
-    "is_parallel": bool,
+    "queue_mng": Optional[str],
 }, total=False)
 Timing = TypedDict('Timing', {
     "name": str,
@@ -190,7 +190,7 @@ JobInfo = TypedDict('JobInfo', {
     "owner": str,
     "company": str,
     "high_priority": bool,
-    "is_parallel": bool,
+    "queue_mng": Optional[str],
     "entry_points": JobInfoEntries,
 })
 CustomCodeResponse = TypedDict('CustomCodeResponse', {
@@ -215,6 +215,16 @@ ModelParamsResponse = TypedDict('ModelParamsResponse', {
 FlushAllQueuesResponse = TypedDict('FlushAllQueuesResponse', {
     "success": bool,
 })
+WorkerScale = TypedDict('WorkerScale', {
+    "success": bool,
+})
+KafkaTopics = TypedDict('KafkaTopics', {
+    "topics": Dict[str, Optional[str]],
+    "create": bool,
+})
+KafkaMessage = TypedDict('KafkaMessage', {
+    "messages": Dict[str, str],
+})
 PutNodeBlob = TypedDict('PutNodeBlob', {
     "key": str,
     "new_uri": str,
